@@ -3,8 +3,6 @@
 # License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).
 
 from __future__ import print_function, unicode_literals
-# from past.builtins import basestring
-
 import os
 import os.path
 import sys
@@ -13,12 +11,12 @@ from string import Template
 from subprocess import Popen, PIPE
 import shutil
 import argparse
-import inspect
 import glob
 from os0 import os0
+import magic
 
 
-__version__ = "1.0.2"
+__version__ = "1.0.5"
 # Module to test version (if supplied version test is executed)
 # REQ_TEST_VERSION = "0.1.4"
 
@@ -49,7 +47,7 @@ OE_CONF = False
 # List of string parameters in [options] of config file
 LX_CFG_S = ('opt_debug', 'opt_verbose', 'opt_noctr')
 # List of pure boolean parameters in [options] of config file
-LX_CFG_B = ('opt_debug', )
+LX_CFG_B = ('opt_debug', 'python2', 'python3')
 # List of string parameters in line command; may be in LX_CFG_S list too
 LX_OPT_CFG_S = ('opt_echo',     'logfn',
                 'opt_tjlib',    'opt_oelib',
@@ -57,7 +55,9 @@ LX_OPT_CFG_S = ('opt_echo',     'logfn',
                 'opt_verbose',  'opt_debug',
                 'opt_noctr',    'run4cover',
                 'max_test',     'min_test',
-                'opt_pattern')
+                'opt_pattern',  'no_run_on_top',
+                'qsanity',      'esanity',
+                'python2',      'python3')
 # List of pure boolean parameters in line command; may be in LX_CFG_S list too
 LX_OPT_CFG_B = ('qsanity', 'esanity', 'opt_debug', 'opt_tjlib', 'opt_oelib')
 # List of numeric parameters in line command; may be in LX_CFG_S list too
@@ -68,7 +68,8 @@ LX_SB = ('dry_run',)
 #
 DEFDCT = {'run4cover': False,
           'opt_debug': False,
-          'opt_new': False}
+          'opt_new': False
+          }
 #
 LX_OPT_ARGS = {'opt_debug': '-B',
                'opt_echo': '-e',
@@ -78,15 +79,17 @@ LX_OPT_ARGS = {'opt_debug': '-B',
                'opt_new': '-N',
                'opt_oelib': '-O',
                'opt_pattern': '-p',
+               'run_on_top': '-R',
                'min_test': '-r',
                'opt_verbose': '-v',
                'max_test': '-z',
                'opt_noctr': '-0',
-               'run4cover': '-1',
+               'run4cover': '-C',
+               'python2': '-2',
                'python3': '-3'}
 
 DEFAULT_COVERARC = r"""# Config file .coveragerc 2019-08-22
-[report]
+[run]
 include =
 #    ${TRAVIS_BUILD_DIR}/*
     *.py
@@ -105,6 +108,7 @@ omit =
     */__init__.py
     */__openerp__.py
     */__manifest__.py
+[report]
 # Regexes for lines to exclude from consideration
 exclude_lines =
     # Have to re-enable the standard pragma
@@ -137,13 +141,13 @@ class Macro(Template):
 
     def substitute(self, **kwargs):
         nk = {}
-        for k,v in kwargs.items():
+        for k, v in kwargs.items():
             nk[k + '}'] = v
         return super(Macro, self).substitute(**nk)
 
     def safe_substitute(self, **kwargs):
         nk = {}
-        for k,v in kwargs.items():
+        for k, v in kwargs.items():
             nk[k + '}'] = v
         return super(Macro, self).safe_substitute(**nk)
 
@@ -531,7 +535,7 @@ class Z0test(object):
         else:
             self.autorun = True
         this_fqn = os.path.abspath(this_fqn or self._get_this_fqn())
-        this = os0.nakedname(os.path.basename(this_fqn))
+        this = os.path.splitext(os.path.basename(this_fqn))[0]
         this_dir = os.getcwd()
         if (not os.path.basename(this_dir) == 'tests' and 
                 not os.path.isdir('./tests')):
@@ -583,14 +587,16 @@ class Z0test(object):
         else:
             self.autorun = True
         self.module_id = id
-        self.pattern = [self.module_id + '_test*', 'test_*']
+        if this == 'zerobug':
+            self.pattern = [self.module_id + '_test*', 'test_*']
+        else:
+            self.pattern = this
         # If auto regression test is executing
         self.def_tlog_fn = os.path.join(self.testdir,
                                         self.module_id + "_test.log")
         self.ctr_list = []
         if self.autorun:
-            self.ctx = self.parseoptest(argv,
-                                        version=version)
+            self.ctx = self.parseoptest(argv, version=version)
             sys.exit(self.main())
 
     def _create_parser(self, version, ctx):
@@ -607,6 +613,7 @@ class Z0test(object):
         -n --dry-run     count and display # unit tests
         -O               load odoorc library (only in bash scripts)
         -q --quiet       run tests without output (quiet mode)
+        -R --run-inner   run test inner mode (no final result)
         -r --restart     restart count next to number
         -s --start       count 1st test next to number (deprecated, use -r)
         -V --version     show version
@@ -616,6 +623,8 @@ class Z0test(object):
         -z --end         display total # tests when execute them
         -0 --no-count    no count # unit tests
         -1 --coverage    run test for coverage (obsoslete)
+        -2               python2
+        -3               python3
         """
         parser = argparse.ArgumentParser(
             description="Regression test on " + self.module_id,
@@ -675,6 +684,10 @@ class Z0test(object):
                             action="store_false",
                             dest="opt_echo_q",
                             default=True)
+        parser.add_argument("-R", "--run-inner",
+                            help="inner mode w/o final messages",
+                            action="store_true",
+                            dest="no_run_on_top")
         parser.add_argument("-r", "--restart",
                             help="set to counted tests, 1st one next to this",
                             dest="min_test",
@@ -715,6 +728,11 @@ class Z0test(object):
                             action="store_true",
                             dest="run4cover",
                             default=True)
+        parser.add_argument("-2", "--python2",
+                            help="use python2",
+                            action="store_true",
+                            dest="python2",
+                            default=False)
         parser.add_argument("-3", "--python3",
                             help="use python3",
                             action="store_true",
@@ -729,9 +747,8 @@ class Z0test(object):
                 ctx.get('min_test', None) is None) and \
                 ('max_test' not in ctx or
                  ctx.get('max_test', None) is None):
-            ctx['run_on_top'] = True
-            # del ctx['min_test']
-            # del ctx['max_test']
+            if 'run_on_top' not in ctx:
+                ctx['run_on_top'] = True
             ctx['min_test'] = 0
             ctx['max_test'] = 0
         else:
@@ -825,6 +842,8 @@ class Z0test(object):
                         ctx[p] = int(getattr(opt_obj, 'min_test2'))
                     else:
                         ctx[p] = None
+                elif p == 'no_run_on_top' and hasattr(opt_obj, p):
+                    ctx['run_on_top'] = not getattr(opt_obj, p)
                 elif hasattr(opt_obj, p):
                     ctx[p] = getattr(opt_obj, p)
             for p in LX_OPT_CFG_B:
@@ -838,25 +857,27 @@ class Z0test(object):
         return ctx
 
     def _get_this_fqn(self):
-        i = 1
-        valid = False
-        auto_this = False
-        while not valid and i < len(inspect.stack()):
-            this_fqn = os.path.abspath(inspect.stack()[i][1])
-            this = os0.nakedname(os.path.basename(this_fqn))
-            if this[0] == '<' and this[-1] == '>':
-                i += 1
-            elif this in ("__init__", "pdb", "cmd", "z0testlib"):
-                i += 1
-                if this == "__init__":
-                    auto_this = this_fqn
-            else:
-                valid = True
-                if this in ('pkgutil', 'runpy'):
-                    this_fqn = os.path.dirname(auto_this)
-                    id = 'test_%s.py' % os.path.basename(this_fqn)
-                    this_fqn = os.path.join(this_fqn, id)
-        return this_fqn
+        # i = 1
+        # valid = False
+        # auto_this = False
+        # this_fqn = False
+        # while not valid and i < len(inspect.stack()):
+        #     this_fqn = os.path.abspath(inspect.stack()[i][1])
+        #     this = os.path.splitext(os.path.basename(this_fqn))[0]
+        #     if this[0] == '<' and this[-1] == '>':
+        #         i += 1
+        #     elif this in ("__init__", "pdb", "cmd", "z0testlib"):
+        #         i += 1
+        #         if this == "__init__":
+        #             auto_this = this_fqn
+        #     else:
+        #         valid = True
+        #         if this in ('pkgutil', 'runpy'):
+        #             this_fqn = os.path.dirname(auto_this)
+        #             id = 'test_%s.py' % os.path.basename(this_fqn)
+        #             this_fqn = os.path.join(this_fqn, id)
+        # return this_fqn
+        return os.path.abspath(sys.argv[0])
 
     def parseoptest(self, arguments, version=None, tlog=None):
         ctx = {}
@@ -865,7 +886,7 @@ class Z0test(object):
         ctx['testdir'] = self.testdir
         this_fqn = self._get_this_fqn()
         ctx['this_fqn'] = this_fqn
-        this = os0.nakedname(os.path.basename(this_fqn))
+        this = os.path.splitext(os.path.basename(this_fqn))[0]
         ctx['this'] = this
         if os.isatty(0):
             ctx['run_daemon'] = False
@@ -894,7 +915,7 @@ class Z0test(object):
         return DEFDCT
 
     def _inherit_opts(self, ctx):
-        args = []
+        args = ['-R']
         for p in LX_OPT_CFG_S:
             if p == 'opt_echo':
                 if p in ctx and ctx[p]:
@@ -908,7 +929,7 @@ class Z0test(object):
                 if p in ctx and ctx[p]:
                     args.append(LX_OPT_ARGS[p] + ctx[p])
             elif p == 'run4cover':
-                if p in ctx and ctx[p]:
+                if p in ctx and not ctx[p]:
                     args.append(LX_OPT_ARGS[p])
             elif p == 'min_test':
                 args.append(LX_OPT_ARGS[p] + str(ctx['ctr']))
@@ -958,6 +979,22 @@ class Z0test(object):
             ctx = self._restore_opt(ctx, p)
         return ctx
 
+    def set_shabang(self, ctx, testfile):
+        if (not ctx.get('python3', False) and
+                not ctx.get('python2', False) and
+                os.path.isfile(testfile)):
+            do_rewrite = False
+            with open(testfile, 'rb') as fd:
+                source = fd.read().decode('utf-8')
+                if source.startswith('#!'):
+                    do_rewrite = True
+                    new_source = '#!%s\n' % sys.executable
+                    for ln in source.split('\n')[1:]:
+                        new_source += '%s\n' % ln
+            if do_rewrite:
+                with open(testfile, 'wb') as fd:
+                    fd.write(new_source.encode('utf-8'))
+
     def test_version(self, ctx, testname):
         """testname format is '__version_T_VVVVFILE' where:
             T: type - May be:
@@ -1017,9 +1054,9 @@ class Z0test(object):
         file = Macro(testname[10:])
         file = file.safe_substitute(**ctx)
         msg = "doctest %s" % os.path.basename(file)
-        # cmd = 'python -m doctest %s' % file
+        # cmd = 'sys.executable -m doctest %s' % file
         FNULL = open(os.devnull, 'w')
-        res = subprocess.call(['python', '-m', 'doctest', file],
+        res = subprocess.call(['sys.executable', '-m', 'doctest', file],
                               stdout=FNULL,
                               stderr=subprocess.STDOUT)
         return self.test_result(ctx,
@@ -1027,30 +1064,19 @@ class Z0test(object):
                                 TEST_SUCCESS,
                                 res)
 
-    def set_mime_python_ver(testname, with_python3):
-        fd = open(testname, 'rbU')
-        code = fd.read()
-        fd.close()
-        code.replace('#!/usr/bin/env python',
-                     '#!/usr/bin/env python3',
-                     1)
-        fd = open(testname, 'w')
-        fd.write(code)
-        fd.close()
-
     def _exec_tests_4_count(self, test_list, ctx, TestCls=None):
         if ctx.get('_run_autotest', False):
             self.dbgmsg(ctx, '>>> exec_tests_4_count(autotest)')
         else:
             self.dbgmsg(ctx, '>>> exec_tests_4_count(%s)' % test_list)
-        opt4childs = ['-n']
+        opt4childs = ['-n', '-R']
         ctx = self._ready_opts(ctx)
         ctx = self._save_options(ctx)
         testctr = 0
         if TestCls:
             T = TestCls(self)
-        if TestCls and hasattr(TestCls, 'setup'):
-            getattr(T, 'setup')(ctx)
+            if hasattr(TestCls, 'setup'):
+                getattr(T, 'setup')(ctx)
         for testname in test_list:
             self.dbgmsg(ctx,
                         '- min=%d, max=%d, ctr=%d, -0=%s, Cover=%s' %
@@ -1063,7 +1089,6 @@ class Z0test(object):
             basetn = os.path.basename(testname)
             ctx['ctr'] = 0
             if testname.startswith('__test'):
-                # ctx['dry_run'] = True
                 ctx['ctr'] = int(testname[7:9])
             elif testname.startswith('__version'):
                 self.test_version(ctx, testname)
@@ -1071,14 +1096,18 @@ class Z0test(object):
                 self.doctest(ctx, testname)
             elif TestCls and hasattr(TestCls, testname):
                 getattr(T, testname)(ctx)
-            elif os0.nakedname(basetn) != ctx['this']:
+            elif os.path.splitext(basetn)[0] != ctx['this']:
+                mime = magic.Magic(
+                    mime=True).from_file(os.path.realpath(testname))
                 if os.path.dirname(testname) == "":
                     testname = os.path.join(self.testdir, testname)
-                if basetn.endswith('.py'):
+                if mime == 'text/x-python':
                     if ctx.get('python3', False):
                         test_w_args = ['python3'] + [testname] + opt4childs
+                    elif ctx.get('python2', False):
+                        test_w_args = ['python2'] + [testname] + opt4childs
                     else:
-                        test_w_args = ['python'] + [testname] + opt4childs
+                        test_w_args = [sys.executable] + [testname] + opt4childs
                 else:
                     test_w_args = [testname] + opt4childs
                 self.dbgmsg(ctx, ">>>  test_w_args=%s" % test_w_args)
@@ -1123,14 +1152,6 @@ class Z0test(object):
         self.dbgmsg(ctx, '- c=%d, ctr_list=%s' % (ctx['ctr'], self.ctr_list))
         if TestCls:
             T = TestCls(self)
-        if (ctx.get('run4cover', False) and
-                not ctx.get('dry_run', False)):
-            try:
-                subprocess.call(['coverage', 'erase'],
-                    stdout=open('/dev/null', 'w'),
-                    stderr=open('/dev/null', 'w'))
-            except OSError:
-                print('Coverage not found!')
         if TestCls and hasattr(TestCls, 'setup'):
             getattr(T, 'setup')(ctx)
         for testname in test_list:
@@ -1156,32 +1177,40 @@ class Z0test(object):
                 if ctx.get('opt_debug', False):
                     self.dbgmsg(ctx, ">>> %s()" % testname)
                 sts = getattr(T, testname)(ctx)
-            elif os0.nakedname(basetn) != ctx['this']:
+            elif os.path.splitext(basetn)[0] != ctx['this']:
+                mime = magic.Magic(
+                    mime=True).from_file(os.path.realpath(testname))
                 if os.path.dirname(testname) == "":
                     testname = os.path.join(self.testdir, testname)
-                if basetn.endswith('.py') or basetn.endswith('.pyc'):
+                if mime == 'text/x-python':
                     self.dbgmsg(ctx, '- ctr=%d' % ctx['ctr'])
                     if os.environ.get('TRAVIS_PDB') == 'true':
                         if ctx.get('python3', False):
                             test_w_args = ['python3', '-m', 'pdb', testname
                                            ] + opt4childs
+                        elif ctx.get('python2', False):
+                            test_w_args = ['python2', '-m', 'pdb', testname
+                                           ] + opt4childs
                         else:
-                            test_w_args = ['python', '-m', 'pdb', testname
+                            test_w_args = [sys.executable, '-m', 'pdb', testname
                                            ] + opt4childs
                     elif (ctx.get('run4cover', False) and
                             not ctx.get('dry_run', False)):
+                        self.set_shabang(ctx, testname)
                         test_w_args = [
                             'coverage',
                             'run',
-                            # '-a',
+                            '-a',
                             '--rcfile=%s' % ctx['COVERAGE_PROCESS_START'],
                             testname
                         ] + opt4childs
                     else:
                         if ctx.get('python3', False):
                             test_w_args = ['python3'] + [testname] + opt4childs
+                        elif ctx.get('python2', False):
+                            test_w_args = ['python2'] + [testname] + opt4childs
                         else:
-                            test_w_args = ['python'] + [testname] + opt4childs
+                            test_w_args = [sys.executable] + [testname] + opt4childs
                     self.dbgmsg(ctx, ">>> subprocess.call(%s)" % test_w_args)
                     try:
                         sts = subprocess.call(test_w_args)
@@ -1194,7 +1223,8 @@ class Z0test(object):
                         sts = subprocess.call(test_w_args)
                     except OSError:
                         sts = 127
-                ctx['ctr'] += self.ctr_list[ix]
+                if not ctx.get('opt_noctr', False):
+                    ctx['ctr'] += self.ctr_list[ix]
                 ix += 1
             if sts or ctx.get('teststs', TEST_SUCCESS):      # pragma: no cover
                 sts = TEST_FAILED
@@ -1214,7 +1244,8 @@ class Z0test(object):
             if hasattr(Test, tname):
                 test_list.append(tname)
             test_num += 1
-        self._exec_tests_4_count(test_list, ctx, Test)
+        if not ctx.get('opt_noctr', False):
+            self._exec_tests_4_count(test_list, ctx, Test)
         if ctx.get('dry_run', False):
             if not ctx.get('_run_autotest', False):
                 print(ctx['max_test'])
@@ -1226,10 +1257,10 @@ class Z0test(object):
                                   echo=ctx.get('opt_echo', False))
             sts = self._exec_all_tests(test_list, ctx, Test)
         return sts
-
-    def main_file(self, ctx=None, Test=None, UT1=None, UT=None):
-        print('Deprecatede method: use main(..)')
-        self.main(ctx=ctx, Test=Test, UT1=UT1, UT=UT)
+    #
+    # def main_file(self, ctx=None, Test=None, UT1=None, UT=None):
+    #     print('Deprecatede method: use main(..)')
+    #     self.main(ctx=ctx, Test=Test, UT1=UT1, UT=UT)
 
     def main(self, ctx=None, Test=None, UT1=None, UT=None):
         """Default main program for test execution
@@ -1256,11 +1287,15 @@ class Z0test(object):
         if (ctx['this'] != 'test_zerobug' and
                 ctx.get('run_on_top', False) and
                 not ctx.get('_run_autotest', False)):
-            # sts = self.sanity_check('-q')
-            sts = 0  #debug
-            if sts == TEST_FAILED:                         # pragma: no cover
-                print("Invalid test library!")
-                exit(sts)
+            if (ctx.get('run4cover', False) and
+                    not ctx.get('dry_run', False)):
+                try:
+                    subprocess.call(['coverage', 'erase'],
+                                    stdout=open('/dev/null', 'w'),
+                                    stderr=open('/dev/null', 'w'))
+                except OSError:
+                    print('Coverage not found!')
+                    ctx['run4cover'] = False
         test_list = []
         if UT is not None and isinstance(UT, list):
             self.dbgmsg(ctx, '>>> test_list=%s' % UT)
@@ -1274,12 +1309,9 @@ class Z0test(object):
                 test_files = os.path.abspath(
                     os.path.join(self.testdir, pattern))
                 for fn in sorted(glob.glob(test_files)):
-                    if len(fn) - fn.rfind('.') <= 4:
-                        if fn.endswith('.py'):
-                            test_list.append(fn)
-                        elif os.access(fn, os.X_OK) and fn.endswith('.sh'):
-                            test_list.append(fn)
-                    elif os.access(fn, os.X_OK) and os.name == 'posix':
+                    mime = magic.Magic(
+                        mime=True).from_file(os.path.realpath(fn))
+                    if mime in ('text/x-python', 'text/x-shellscript'):
                         test_list.append(fn)
         if len(test_list) == 0 and Test is not None:
             self.dbgmsg(ctx, '- len(test_list) == 0 ')
@@ -1398,7 +1430,7 @@ class Z0test(object):
         sts = self.main(z0ctx, SanityTest)
         if full:
             for p in 'min_test', 'max_test', 'ctr':
-                    full[p] = z0ctx[p]
+                full[p] = z0ctx[p]
         del z0ctx
         return sts
 
@@ -1409,6 +1441,8 @@ class Z0test(object):
             os.path.dirname(ctx.get('this_fqn', './Z0BUG/tests')), 'res')
         if not os.path.isdir(root):
             os.mkdir(root)
+        if not isinstance(os_tree, (list, tuple)):
+            os_tree = [os_tree]
         for path in os_tree:
             if path[0] not in ('~', '/') and not path.startswith('./'):
                 path = os.path.join(root, path)
@@ -1421,6 +1455,8 @@ class Z0test(object):
         root = os.path.join(os.path.dirname(ctx['this_fqn']), 'res')
         if not os.path.isdir(root):
             return
+        if not isinstance(os_tree, (list, tuple)):
+            os_tree = [os_tree]
         for path in os_tree:
             if path[0] not in ('.', '/'):
                 path = os.path.join(root, path)
@@ -1431,29 +1467,107 @@ class Z0test(object):
 
 class Z0testOdoo(object):
 
-    def build_odoo_env(self, ctx, version, hierarchy=None):
+    def __init__(self, argv=None):
+        this_fqn = None
+        if not argv and len(sys.argv) and not sys.argv[0].startswith('-'):
+            this_fqn = sys.argv[0]
+            this_fqn = os.path.abspath(this_fqn)
+        this_dir = os.getcwd()
+        if (not os.path.basename(this_dir) == 'tests' and
+                not os.path.isdir('./tests')):
+            this_dir = os.path.dirname(this_fqn)
+        self.this_dir = this_dir
+        if os.path.basename(this_dir) == 'tests':
+            self.testdir = self.this_dir
+            self.rundir = os.path.abspath(os.path.join(self.this_dir, '..'))
+        else:                                               # pragma: no cover
+            if os.path.isdir('./tests'):
+                self.testdir = os.path.join(self.this_dir,
+                                            'tests')
+            else:
+                self.testdir = self.this_dir
+            self.rundir = self.this_dir
+
+    def get_outer_dir(self):
+        """Get dir out of current virtual environment
+        In local tests it serves to find local Odoo repositories to avoid
+        git clone or wget from web.
+        """
+        outer_dir = os.environ.get('TRAVIS_BUILD_DIR', os.getcwd())
+        if os.path.basename(outer_dir) == 'tests':
+            outer_dir = os.path.abspath(
+                os.path.join(outer_dir, '..', '..', '..', '..'))
+        else:
+            outer_dir = os.path.abspath(
+                os.path.join(outer_dir, '..', '..', '..'))
+        return outer_dir
+
+    def get_local_odoo_path(self, git_org, reponame, branch, home=None):
+        outer_dir = home or self.get_outer_dir()
+        majver = branch.split('.')[0]
+        found_path = False
+        for reporg in (reponame, reponame.lower(), git_org, git_org.lower()):
+            if found_path:
+                break
+            for odoo_ver in (branch, majver):
+                if found_path:
+                    break
+                src_repo_path = os.path.join(
+                    outer_dir, '%s%s' % (reporg, odoo_ver))
+                if os.path.isdir(src_repo_path):
+                    found_path = True
+                    break
+                src_repo_path = os.path.join(
+                    outer_dir, '%s-%s' % (reporg, odoo_ver))
+                if os.path.isdir(src_repo_path):
+                    found_path = True
+                    break
+                if reponame == 'OCB':
+                    continue
+                for nm in ('', 'extra', 'private-addons', 'powerp'):
+                    if nm:
+                        src_repo_path = os.path.join(
+                            outer_dir, odoo_ver, nm, reporg)
+                    else:
+                        src_repo_path = os.path.join(outer_dir, odoo_ver, reporg)
+                    if os.path.isdir(src_repo_path):
+                        found_path = True
+                        break
+        if not found_path:
+            # Local dir of current project is like '~/12.0'
+            src_repo_path = os.path.join(outer_dir, branch)
+            if not os.path.isdir(src_repo_path):
+                src_repo_path = False
+        return src_repo_path
+
+    def build_odoo_env(self, ctx, version,
+                       hierarchy=None, name=None, retodoodir=None):
         """Build a simplified Odoo directory tree
-        version: 14.0, 13.0, ..., 7.0, 6.1
+        version: 15.0, 14.0, 13.0, ..., 7.0, 6.1
+        name: name of odoo dir (default equal to version)
         hierarchy: flat,tree,server (def=flat)
         """
-        if version in ('10.0', '11.0', '12.0', '13.0', '14.0'):
+        name = name or version
+        if version in ('10.0', '11.0', '12.0', '13.0', '14.0', '15.0'):
             if hierarchy == 'tree':
-                odoo_home = os.path.join(version, 'odoo', 'odoo')
+                odoo_home = os.path.join(name, 'odoo', 'odoo')
             else:
-                odoo_home = os.path.join(version, 'odoo')
+                odoo_home = os.path.join(name, 'odoo')
             script = 'odoo-bin'
         elif version in ('6.1', '7.0', '8.0', '9.0'):
             if hierarchy == 'server':
-                odoo_home = os.path.join(version, 'server', 'openerp')
+                odoo_home = os.path.join(name, 'server', 'openerp')
             else:
-                odoo_home = os.path.join(version, 'openerp')
+                odoo_home = os.path.join(name, 'openerp')
             script = 'openerp-server'
         else:
             raise KeyError('Invalid Odoo version')
-        os_tree = [version,
-                   os.path.join(version, 'addons'),
+        os_tree = [name,
+                   os.path.join(name, 'addons'),
                    odoo_home,
-                   os.path.join(odoo_home, 'addons')]
+                   os.path.join(odoo_home, 'addons'),
+                   os.path.join(name, '.git'),
+                   ]
         root = Z0test().build_os_tree(ctx, os_tree)
         RELEASE_PY = '''
 RELEASE_LEVELS = [ALPHA, BETA, RELEASE_CANDIDATE, FINAL] = ['alpha', 'beta', 'candidate', 'final']
@@ -1464,15 +1578,88 @@ RELEASE_LEVELS_DISPLAY = {ALPHA: ALPHA,
 version_info = (%s, %s, 0, 'final', 0, '')
 version = '.'.join(map(str, version_info[:2])) + RELEASE_LEVELS_DISPLAY[version_info[3]] + str(version_info[4] or '') + version_info[5]
 series = serie = major_version = '.'.join(map(str, version_info[:2]))'''
-        with open(os.path.join(root, odoo_home, 'release.py'), 'w') as fd:
+        if name[0] not in ('~', '/') and not name.startswith('./'):
+            odoo_root = os.path.join(root, name)
+        else:
+            odoo_root = name
+        odoo_home = os.path.join(os.path.dirname(odoo_root), odoo_home)
+        with open(os.path.join(odoo_home, 'release.py'), 'w') as fd:
             versions = version.split('.')
             fd.write(RELEASE_PY % (versions[0], versions[1]))
-        with open(os.path.join(root, odoo_home, '__init__.py'), 'w') as fd:
+        with open(os.path.join(odoo_home, '__init__.py'), 'w') as fd:
             fd.write('import release\n')
-        with open(os.path.join(root, version, script), 'w') as fd:
+        with open(os.path.join(odoo_root, script), 'w') as fd:
             fd.write('\n')
+        with open(os.path.join(odoo_root, '.travis.yml'), 'w') as fd:
+            fd.write('\n')
+        with open(os.path.join(odoo_root, 'README.rst'), 'w') as fd:
+            fd.write('\n')
+        if retodoodir:
+            return odoo_root
         return root
 
+    def create_repo(self, ctx, root, reponame, version,
+                    hierarchy=None, name=None, repotype=None):
+        REPOTYPES = {
+            'oca': {
+                'd': ['.git',],
+                'f': ['README.md', '.travis.yml'],
+            },
+            'zero': {
+                'd': ['egg-info', '.git'],
+                'f': ['README.rst', '.travis.yml'],
+            },
+        }
+        name = name or version
+        repotype = repotype or 'oca'
+        odoo_root = os.path.join(root, name)
+        if not os.path.isdir(odoo_root):
+            odoo_root = root
+        repodir = os.path.join(odoo_root, reponame)
+        if not os.path.isdir(repodir):
+            os.mkdir(repodir)
+        for ldir in REPOTYPES.get(repotype, {}).get('d', []):
+            path = os.path.join(repodir, ldir)
+            if not os.path.isdir(path):
+                os.mkdir(path)
+        for fn in REPOTYPES.get(repotype, {}).get('f', []):
+            path = os.path.join(repodir, fn)
+            if not os.path.isfile(path):
+                open(path, 'w').close()
+        return repodir
+
+    def create_module(self, ctx, repo_root, name, version, moduletype=None):
+        MODULETYPES = {
+            'simple': {
+                'd': [],
+                'f': ['__init__.py'],
+            },
+        }
+        moduletype = moduletype or 'simple'
+        moduledir = os.path.join(repo_root, name)
+        if not os.path.isdir(moduledir):
+            os.mkdir(moduledir)
+        majver = int(version.split('.')[0])
+        if majver < 10:
+            ffn = os.path.join(moduledir, '__openerp__.py')
+        else:
+            ffn = os.path.join(moduledir, '__manifest__.py')
+        with open(ffn, 'w') as fd:
+            fd.write(str({
+                'name': name,
+                'version': version,
+                'summary': 'This module is a fake, just for test!',
+            }))
+        for ldir in MODULETYPES.get(moduletype, {}).get('d', []):
+            path = os.path.join(moduledir, ldir)
+            if not os.path.isdir(path):
+                os.mkdir(path)
+        for fn in MODULETYPES.get(moduletype, {}).get('f', []):
+            path = os.path.join(moduledir, fn)
+            if not os.path.isfile(path):
+                open(path, 'w').close()
+        return moduledir
+
     def real_git_clone(self, remote, reponame, branch, odoo_path):
         odoo_url = 'https://github.com/%s/%s.git' % (remote, reponame)
         os.system('git clone --depth=50 %s -b %s %s' % (
@@ -1485,30 +1672,26 @@ series = serie = major_version = '.'.join(map(str, version_info[:2]))'''
         if force or os.environ.get('TRAVIS') == 'true':
             self.real_git_clone(remote, reponame, branch, odoo_path)
         elif not os.path.isdir(odoo_path):
-            majver = branch.split('.')[0]
-            if remote == 'OCA':
-                src_rep_path = os.path.join(
-                    os.environ.get('TRAVIS_SAVED_HOME', os.environ['HOME']),
-                    '%s%s' % (remote.lower(), majver))
-                if not os.path.isdir(src_rep_path):
-                    self.real_git_clone(remote, reponame, branch, odoo_path)
-                    os.rename(odoo_path, src_rep_path)
-                if not os.path.isdir(src_rep_path):
-                    src_rep_path = None
-            else:
-                src_rep_path = os.path.join(
-                    os.environ.get('TRAVIS_SAVED_HOME', os.environ['HOME']),
-                    branch)
-            if reponame == 'OCB':
+            src_repo_path = self.get_local_odoo_path(remote, reponame, branch)
+            if not os.path.isdir(src_repo_path):
+                self.real_git_clone(remote, reponame, branch, odoo_path)
+            elif reponame == 'OCB':
                 for nm in ('addons', 'odoo', 'openerp'):
-                    src_path = os.path.join(src_rep_path, nm)
+                    src_path = os.path.join(src_repo_path, nm)
                     dst_path = os.path.join(odoo_path, nm)
                     if os.path.isdir(src_path):
-                        shutil.copytree(src_path, dst_path)
+                        shutil.copytree(
+                            src_path, dst_path,
+                            symlinks=True,
+                            ignore=shutil.ignore_patterns(
+                                '*.pyc', '.idea/', 'setup/'))
                 for nm in ('.travis.yml', 'odoo-bin', 'openerp-server',
                            'openerp-wsgi.py', 'requirements.txt'):
-                    src_path = os.path.join(src_rep_path, nm)
+                    src_path = os.path.join(src_repo_path, nm)
                     if os.path.isfile(src_path):
                         shutil.copy(src_path, odoo_path)
             else:
-                shutil.copytree(src_rep_path, odoo_path)
+                shutil.copytree(
+                    src_repo_path, os.path.join(odoo_path, reponame),
+                    symlinks=True,
+                    ignore=shutil.ignore_patterns('*.pyc', '.idea/', 'setup/'))
\ No newline at end of file
